-- (c) IPM, University of Minho (2008/09 ; 2009/10)
-- Version 1.0

module Abc where

import Cp
import Ratio
import Data.Char
import Data.List
import System.Cmd
import Data.HashTable
import Ipm

class Abc a
      where abc :: a -> String

data Pitch = Pi(Integer, Note) deriving Show

instance Abc Pitch
      where abc (Pi x) = n2abc x

instance Abc Note  -- from Ipm
      where abc = show

instance (Abc a) => Abc (P2 a)  -- from Ipm
      where -- abc (P l) = abc l
            abc (m :|: n) = "%%staves [1 2]\nV:1 clef=treble\nV:2 clef=treble\n" ++
                            "[V:1] "++(abc m) ++ "\n[V:2] "++ (abc n) ++ "\n"

instance (Integral a) => Abc (Ratio a)
      where abc r = show(numerator r) ++ "/" ++ show(denominator r)

instance (Abc a, Abc b) => Abc(a,b) where
    abc(a,b) = (abc a)++(abc b)

instance (Abc a) => Abc [a] where
    abc = concat . (map abc) 

instance (Abc a) => Abc(N a) where -- from Ipm
    abc (N a) = abc a
    abc Bar = "|"
    abc TBar = "|-"
    abc Tie = "-"
    abc BTie = "("
    abc ETie = ")"

instance Abc Char where
    abc c = [c]

play k m a =
--- a is already an (abc) string
--- k: Key
--- m: meter
       let s = ("\nX:1\nM:"++m++"\nL:1/1\nK:"++k++"\n"++ a)
           fn = "_showplay_"++show(hashString s)
           pdf = fn++".pdf"
       in do { writeFile "_.abc" s;
	       system $ "abcm2ps -c _.abc; pstopdf -o "++pdf++" Out.ps";
	       system $ "open "++pdf;
	       system "abc2midi _.abc";
               system "timidity _1.mid"
             }

-- Windows
wplay k m a =
--- a is already an (abc) string
--- k: Key
--- m: meter
       let fn = "_showplay_"++show(hashString s)
           s = ("X:1\nT:"++"play(abc)"++"\nK:"++k++"\nM:"++m++"\nL:1/1\n"++ a)
           pdf = fn++".pdf"
           ps = fn++".ps"
       in do { writeFile "_.abc" s;
	       system $ "abcm2ps -O " ++ ps ++ " _.abc";
	       system $ "ps2pdf " ++ ps;
	       system $ "del " ++ ps ;
	       system   "abc2midi _.abc";
               system   "start /b timidity _1.mid";
 	       system $ "start /b AcroRd32 " ++ pdf
             }

abcPlay k m x
--- x is of class Abc
       = play (abc k) (abc m) (abc (divide (repeat (getmeter m)) x))
         where getmeter "C" = 4%4
               getmeter "C|" = 2%2
               getmeter (c:(_:[d])) = (getnat c) % (getnat d)
               getmeter _ = 4%4
               getnat c = ord c - ord '0'

abcPlay_ k m x
--- x is of class Abc
--- does not divide x into bars according to m
       = play (abc k) (abc m) (abc x)

-- interface between Note and ABC
n2abc :: (Num a) => (a, Note) -> String
n2abc (-2,n)=(note2abc n):",,"
n2abc (-1,n)=(note2abc n):","
n2abc (0,n)=[note2abc n]
n2abc (1,n)=[lowercase (note2abc n)]
n2abc (2,n)=(lowercase (note2abc n)):"'"
n2abc (3,n)=(lowercase (note2abc n)):"''"

note2abc A='A'
note2abc B='B'
note2abc n=chr(ord 'C'+fromEnum n) 
 
lowercase c = chr(ord c + ord 'a' - ord 'A')
uppercase c = chr(ord c - ord 'a' + ord 'A')

