
-- (c) IPM, University of Minho (2008/09 ; 2009/10)
-- Version 1.0

module Ipm where

import Cp
import Ratio
import Data.Char
import Data.List

data P2 a = [a] :|: [a] deriving Show	-- two part music

instance Functor P2
   where fmap f (l :|: r) = (map f l) :|: (map f r) 

-- (0) Bars

class Div t 
      where divide :: (Ord b, Num b) => [b] -> t (a, b) -> t(N (a, b))

instance Div []
   where divide = dividl

instance Div P2
   where divide r (a :|: b) = (divide r a) :|: (divide r b)

-- (1) Augmentation, diminution

class Scalable t 
   where scale :: (Num a, Ord a) => a -> t -> t 

instance (Scalable b) => Scalable (P2 b)
   where scale i (a :|: b) = (scale i a) :|: (scale i b)

instance  (Scalable b, Scalable c) => Scalable (b,c)
   where scale a = (scale a) >< (scale a)

instance (Scalable a) => Scalable [a]
   where scale i = map (scale i)

instance Scalable Char
   where scale i = id

instance (Integral a) => Scalable (Ratio a)
    where scale i x | i < 1 =  x * (1%2)
                    | otherwise = x * 2

-- (2) Pattern matching

findPattern :: (Eq b) => [b] -> [b] -> [Int]
findPattern p l = [ i | i <-inds l , (match p l i) ]
 
match p l i = p `isPrefixOf` (drop i l)

-- repeat rythmic cell forever: Data.List cycle function

-- repeat :: forall a. a -> [a] available from GHC.List

ntimes :: (Integral t) => [a] -> t -> [a]
ntimes x 0     = []
ntimes x (n+1) = x ++ (ntimes x n)

twice = flip ntimes 2

-- abstract from repeated notes
--nrep [] = []
--nrep ((n,d):l) = aux n d l
--             where aux n d [] = [(n,d)]
--                   aux n d ((n',d'):l)
--                        | n==n' = aux n (d+d') l
--                        | n/=n' = (n,d):(aux n' d' l)
nrep [] = []
nrep [x] = [x]
nrep ((n,d):(n',d'):l)
     | n==n' = nrep ((n,d+d'):l)
     | n!=n' = (n,d):nrep((n',d'):l)

-- (2) Sampling for analysis

sample :: (Ord d, Num d) => [d] -> [(n, d)] -> [(n, d)]
sample [] _ = []
sample _ [] = []
sample (y:r) ((a,x):t)
	| y < 0 && x + y == 0  = sample r t
	| y < 0 && x + y >  0  = sample r ((a,x+y):t)
	| y < 0 && x + y <  0  = sample ((x+y):r) t
	| y > 0 && y < x       = (a,y) : sample r ((a,x-y):t)
	| y > 0 && y > x       = (a,y) : sample ((x-y):r) t
	| y > 0 && y == x      = (a,y) : sample r t

-- (3) Introducing bars

data N a = N a | Bar | TBar | Tie | BTie | ETie deriving Show

dividl :: (Ord b, Num b) => [b] -> [(a, b)] -> [N (a, b)]
dividl [] _ = []
dividl _ [] = []
dividl (y:r) ((a,x):t)
 	| y < x       = [ N (a,y) , TBar ] ++ dividl r ((a,x-y):t)
 	| y > x       = ( N (a,x)) : dividl ((y-x):r) t
 	| y == x      = [ N (a,x) , Bar ] ++ dividl r t
 
-- (4) Pitch

data Note = A | B | C | D | E | F | G deriving (Show, Eq)

instance Enum Note
   where succ A = B
         succ B = C
         succ C = D
         succ D = E
         succ E = F
         succ F = G
         succ G = A
	 pred C = B
	 pred n = toEnum(fromEnum n -1)
         fromEnum C = 0
         fromEnum D = 1
         fromEnum E = 2
         fromEnum F = 3
         fromEnum G = 4
         fromEnum A = 5
         fromEnum B = 6
         toEnum 0 = C
         toEnum 1 = D
         toEnum 2 = E
         toEnum 3 = F
         toEnum 4 = G
         toEnum 5 = A
         toEnum 6 = B

instance Ord Note
     where n <= m = fromEnum n <= fromEnum m

next :: (Enum a ) => (a, Note) -> (a, Note)
next = nxt C
       where nxt b (m,n) = (if (succ n) == b then succ m else m,succ n)

previous :: (Enum a ) => (a, Note) -> (a, Note)
previous = prv C
       where prv b (m,n) = (if n == b then pred m else m,pred n)

shift :: (Num a, Enum o, Ord a) => a -> (o, Note) -> (o, Note)
shift 0 p = p 
shift n p | n > 0 = next (shift (n-1) p) 
          | n < 0 = previous (shift (n+1) p) 

ndiff :: (Enum a, Num b, Ord a) => (a, Note) -> (a, Note) -> b
ndiff n m | n >  m    =   aux n m
          | otherwise = -(aux m n)
          where aux n m | n == m = 0
                        | n > m  =1  + aux n (next m)

interval :: (Num b, Ord b, Enum a, Ord a) => (a, Note) -> (a, Note) -> b
interval n m = if d < 0 then d-1 else d+1
               where d = ndiff n m

-- (5) Generic derivative / integral

gderiv :: a -> (a -> a -> c) -> [a] -> [c]
gderiv z f s = ttl (zipWith f s (z:s))
              where ttl [] = []
                    ttl (a:t) = t

ginteg :: (a -> b -> b) -> b -> [a] -> [b]
ginteg f i s = i:aux i s
      where aux i [] = []
            aux i (h:t) = i':(aux i' t)
                          where i'=f h i

-- (6) Note integral

-- integ :: (Num a, Enum o, Ord a) => (o, Note) -> [a] -> [(o, Note)]
integ i s = ginteg shift i s

-- numeric derivative

deriv = gderiv 0 (-)

-- diat b a = (b-a) + if b < a then -1 else 1

-- (7) Auxiliary

cut :: Int -> Int -> [a] -> [a]
cut n m = (take (m-n+1)) . drop (n-1)

inds :: [b] -> [Int]
inds = (map fst) . (zip [1..])

elems :: (Ord a) => [a] -> [a]
elems = sort . nub

-- histogram given population list
hist l = nub [ (x, count x l) | x <- l ]
         where count a l = length [ x | x <- l, x == a]

-- ndcopy is the nub function
ndcopy [] = []
ndcopy [x] = [x]
ndcopy (x:r) = x:(filter (/=x)(ndcopy r))

-- ncdcopy: no consecutive duplicates
ncdcopy [] = []
ncdcopy [x] = [x]
ncdcopy (x:y:l)
     | x==y =   ncdcopy(x:l)
     | x/=y = x:ncdcopy(y:l)

gather l x = map (flip ($) x) l

-- matching

patternIndices p l = [ (i,i+length p-1) | (x,i) <- zip l [0..], match p l i ]
 
sing a = [a]
dup a = [a,a]

a != b = not(a == b)

a |-> b = (a,b)

supermap f [] = []
supermap f (a:l) = a:(map f (supermap f l))

--- list of pairs to list
lp2l :: [(b, b)] -> [b]
lp2l = (concat . map f) where f(a,b)=[a,b]

name2music w = map (g.f) w
	where g c = [c]
f ' ' = 'z'
f 'H' = 'B'
f 'h' = 'b'
f c | c `elem` ['A'..'Z']++['a'..'z'] = chr(rem(ord c - b) 8 + b)
    | otherwise = 'z'
      where  b = if c `elem` ['A'..'Z'] then ord('A') else ord('a')

notVowel c = not(c `elem` "aeiouAEIOU")

